from base_st_classes import *

class LinkSuite(NFSSuite):
    """Test operation 11: LINK

    FIXME: Add attribute directory and named attribute testing.
    FIXME: More combinations of invalid filehandle types. 

    Equivalence partitioning:

    Input Condition: saved filehandle
        Valid equivalence classes:
            file(1)
            link(2)
            block(3)
            char(4)
            socket(5)
            FIFO(6)
        Invalid equivalence classes:
            dir(7)
            invalid filehandle(8)
    Input Condition: current filehandle
        Valid equivalence classes:
            dir(9)
        Invalid equivalence classes:
            not dir(10)
            invalid filehandle(11)
    Input Condition: newname
        Valid equivalence classes:
            valid name(12)
        Invalid equivalence classes:
            zerolength(13)
            non-utf8(14)

    Comments: It's not possible to cover eq. class 11, since saving a filehandle
    gives a current filehandle as well. 
    """

    def setUp(self):
        NFSSuite.setUp(self)
        self.obj_name = "link1"

        self.lookup_dir_ops = self.ncl.lookup_path(self.tmp_dir)

    def tearDown(self):
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass

    def _prepare_operation(self, file):
        # Put root FH
        ops = [self.ncl.putrootfh_op()]

        # Lookup source and save FH
        ops.extend(self.ncl.lookup_path(file))
        ops.append(self.ncl.savefh_op())

        # Lookup target directory
        ops.append(self.ncl.putrootfh_op())
        ops.extend(self.lookup_dir_ops)

        return ops

    def _link(self,targetfile,linkname,error=[NFS4_OK]):
        """ Try to create a link to target file
        """
        ops = self._prepare_operation(targetfile)
        ops.append(self.ncl.link_op(linkname))
        res = self.ncl.do_ops(ops)
        self.assert_status(res,error)
        return res
    
    #
    # Testcases covering valid equivalence classes.
    #
    def testFile(self):
        """LINK a regular file

        Covered valid equivalence classes: 1, 9, 12
        """
        # FRED - check numlink attr before and after
        self._link(self.regfile,self.obj_name)

    def testLink(self):
        """LINK a symbolic link should succeed or return NFS4ERR_NOTSUPP

        Covered valid equivalence classes: 2, 9, 12
        """
        res = self._link(self.linkfile, self.obj_name, error=[NFS4_OK, NFS4ERR_NOTSUPP])
        if res.status == NFS4ERR_NOTSUPP:
            self.info_message("LINK a symbolic link is not supported")

    def testBlock(self):
        """LINK a block device

        Covered valid equivalence classes: 3, 9, 12
        """
        self._link(self.blockfile, self.obj_name)

    def testChar(self):
        """LINK a character device

        Covered valid equivalence classes: 4, 9, 12
        """
        self._link(self.charfile, self.obj_name)

    def testSocket(self):
        """LINK a socket

        Covered valid equivalence classes: 5, 9, 12
        """
        self._link(self.socketfile, self.obj_name)
    
    def testFIFO(self):
        """LINK a FIFO

        Covered valid equivalence classes: 6, 9, 12
        """
        self._link(self.fifofile, self.obj_name)

    #
    # Testcases covering invalid equivalence classes.
    #

    def testDir(self):
        """LINK a directory should fail with NFS4ERR_ISDIR

        Covered invalid equivalence classes: 7
        """
        self._link(self.dirfile, self.obj_name, error=[NFS4ERR_ISDIR])

    def testNoSfh(self):
        """LINK should fail with NFS4ERR_NOFILEHANDLE if no (sfh)

        Covered invalid equivalence classes: 8

        Comments: LINK should fail with NFS4ERR_NOFILEHANDLE if no
        saved filehandle exists. 
        """
        linkop = self.ncl.link_op(self.obj_name)
        res = self.ncl.do_ops([self.ncl.putrootfh_op(), linkop])
        self.assert_status(res, [NFS4ERR_NOFILEHANDLE])

    def testCfhNotDir(self):
        """LINK should fail with NFS4ERR_NOTDIR if cfh is not dir

        Covered invalid equivalence classes: 10
        """
        # Put root FH
        ops = [self.ncl.putrootfh_op()]

        # Lookup source and save FH
        ops.extend(self.ncl.lookup_path(self.regfile))
        ops.append(self.ncl.savefh_op())

        # Lookup target directory (a file, this time)
        ops.append(self.ncl.putrootfh_op())
        ops.extend(self.ncl.lookup_path(self.regfile))

        # Link operation
        linkop = self.ncl.link_op(self.obj_name)
        ops.append(linkop)
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_NOTDIR])

    def testZeroLengthName(self):
        """LINK with zero length new name should fail with NFS4ERR_INVAL

        Covered invalid equivalence classes: 13
        """
        self._link(self.regfile, "", error=[NFS4ERR_INVAL])

    def testNonUTF8(self):
        """LINK with non-UTF8 name should return NFS4ERR_INVAL

        Covered valid equivalence classes: 14
        """
        for name in self.get_invalid_utf8strings():
            self._link(self.regfile, name, error=[NFS4ERR_INVAL])

    #
    # Extra tests.
    #
    def testDots(self):
        """LINK with newname . or .. should succeed or return NFS4ERR_INVAL

        Extra test

        Comments: Servers supporting . and .. in file names should
        return NFS4_OK. Others should return
        NFS4ERR_INVAL. NFS4ERR_EXIST should not be returned.
        """
        self._link(self.regfile, ".", error=[NFS4_OK,NFS4ERR_INVAL])
        self._link(self.regfile, "..", error=[NFS4_OK,NFS4ERR_INVAL])

    def testNamingPolicy(self):
        """LINK should obey OPEN file name creation policy

        Extra test
        """
        self.init_connection()

        try:
            (x, rejected_names_open) = self.try_file_names(creator=self.create_via_open)
            (x, rejected_names_link) = self.try_file_names(creator=self.create_via_link)
            self.failIf(rejected_names_open != rejected_names_link,
                        "LINK does not obey OPEN naming policy")
        except SkipException, e:
            self.skip(e)

    #
    # Testing various errors
    #
    def testNoCfh(self):
        """LINK should fail with NFS4ERR_NOFILEHANDLE if no (cfh)

        Note that no (cfh) means no (sfh)
        """
        # Link operation
        linkop = self.ncl.link_op(self.obj_name)
        res = self.ncl.do_ops([linkop])
        self.assert_status(res,[NFS4ERR_NOFILEHANDLE])
      
    def testLongName(self):
        """LINK should fail with NFS4ERR_NAMETOOLONG with long filenames
        """
        longname = "a"*512
        self._link(self.regfile, longname, error=[NFS4ERR_NAMETOOLONG])
        
    def testExists(self):
        """LINK should fail with NFS4ERR_EXIST if file exists
        """

        # Make sure link file exists
        fh = self.do_rpc(self.ncl.do_getfh,self.tmp_dir)
        entries = self.do_rpc(self.ncl.do_readdir,fh)
        names = [entry.name for entry in entries]
        if self.obj_name not in names:
            self._link(self.regfile, self.obj_name)

        # Now try to create it again
        ops = self._prepare_operation(self.regfile)
        linkop = self.ncl.link_op(self.obj_name)
        ops.append(linkop)
        res = self.ncl.do_ops(ops)
        self.assert_status(res,[NFS4ERR_EXIST])

